unit MDPluginAPI;
{**********************************************************
  Multidec to DVBCore API translator.
  Authors: SatRookie and Saar

  Important:
    This is OpenSource project. You can not use this
    source or any idea You got from this source in
    'not-publicly available' or in 'ClosedSource' software!

  Revision history:
    Ver 1.27 October 4, 2004
        packet size for plugins is now tunable (Scribitus)

    Ver 1.26e July 10, 2004
        fixes in GetCurCAT (Scribitus)

    Ver 1.26d December 23, 2003
        fix in TTransponder structure (Scribitus)

    Ver 1.26c October 14, 2003
        fixed bug: save options to ini-file (for MT3.03) (Scribitus)

    Ver 1.26b October 14, 2003
        small fixes (if DLL not MD plugin) (Scribitus)

    Ver 1.26a September 11, 2003
        recompiled with new DVBCore SDK (fix for MD Fenrir) (tnx to Saar)
                                                                   (Scribitus)
    Ver 1.26 September 10, 2003
        fixed problem with unloading plugins and wrapper also (Scribitus)

    Ver 1.25 June 20, 2003
        no more modal dialogs
        corrections in TPInfo<->TProgramm convertation
                                                                   (Scribitus)

    Ver 1.24a June 8, 2003
        corrections in MDAPI_START_FILTER and MDAPI_STOP_FILTER
                                                                   (Scribitus)

    Ver 1.24 June 4, 2003
        Windows API functions for allocate and release buffers
        corrections in DVBCore2MDFilterThunk (some optimizations)
        corrections in TPInfo->TProgramm convertation (check for MAX_CA_SYSTEMS)
                                                                   (Scribitus)

    Ver 1.23 May 29, 2003
        corrected calculation of DiSeq value
        corrections in PROGAPI_GET_CHANNEL_NAME and PROGAPI_SET_CHANNEL_NAME
        corrected buffer releasing
        added check of buffer existence in DVBCore2MDFilterThunk
        some detection of Plugins with MD-API 3.0 (not the best)
                                                                    (Scribitus)
    Ver 1.22 May 22, 2003
        added delay in MDAPI_START_FILTER, need for some Plugins
        corrections in DVBCore2MDFilterThunk, now all MD Plugins work correctly
        corrections in Plugins initialization (when Plugin not able or fail to load)
        corrections in Plugins unloading
        added support for MDAPI_RESCAN_PROGRAMM (emulation)
        added support for MDAPI_GET_TRANSPONDER
        added support for MDAPI_SET_TRANSPONDER
                                                                    (Scribitus)
    Ver 1.21 May 19, 2003
        switch to new exported functions (Saar)
        all MD Plugins now unloaded on DVBCore call DVBPLGClose (Scribitus)

    Ver 1.20 May 17, 2003
        corrected small bug with EMM in GetCurrentChannelMDInfo
        corrections in CloseFilters buffers now be released
        corrections in Plugins initialization
        corrections in unloading Plugins
        added notify call to On_Filter_Close in CloseFilters
        added support for MDAPI_GET_PROGRAMM_NUMMER
        added support for MDAPI_SET_PROGRAMM_NUMMER
        added support for MDAPI_SET_PROGRAMM
        added support for MDAPI_GET_SIGNAL_INFO
        added support for PROGAPI_GET_SIGNAL_LEVEL
        added support for MDAPI_SAVE_PROGRAMM (save in memory for one channel only)
        MDAPI version increased to 3.0 (MD 9.x)
                                                                    (Scribitus)
    Ver 1.10d May 14, 2003
        added 1kB buffer in DVBCore2MDFilterThunk for compatibility with MDAPI
                                                                    (Scribitus)

    Ver 1.10c April 29, 2003
        corrections in StayOnTop mode (Saar)

    Ver 1.10 April 9, 2003
        corrections in Plugins initialization. (Saar)

    Ver 1.09 April 2, 2003
        changes in CK logging. Now displays keys from all SoftCAMs
        which even not uses Wrapper. (Saar)

    Ver 1.08c January 16, 2003
        corrected Audio PID selection to comply
        with DVBCore >=v1.67   (Saar)

    Ver 1.08 January 11, 2003
        Added option "SCAN_CURRENT_CAT as SECTION Filter".
        Some plugins need it on transponders with very low traffic from PID 1.
                                                                    (Saar)

    Ver 1.07 January 02, 2003
        Translation DVBCore FilterID -> MD FilterID
        need for some MD plugins. (Saar)

    Ver 1.06 Desember 15, 2002
        AC3 og Teletext arbeider naa. (saar)

    Ver 1.05 Nevember 15, 2002
        bugfixes in MDAPI_SCAN_CURRENT_CAT/MDAPI_GET_PROGRAMM. (saar)

    Ver 1.04 Nevember 15, 2002
        Correct emulation SCAN_CURRENT_CAT. (saar)

    Ver 1.03 Nevember 15, 2002

    Ver 1.02 Nevember 12, 2002

    Ver 1.01 October 24, 2002
      for actual version of DVBCore (saar)

    Ver 1.0 October 12, 2002
      Now it's separate module   (saar)

    Ver 0.9 October 03, 2002
      Initial release   (SatRookie)


***********************************************************}
{$A+}
interface

uses   Windows, Messages, SysUtils, Variants, Classes, Controls, Forms,
  Dialogs, Menus, DVBCore, StdCtrls, IniFiles{, Graphics};


function DVBPLGInit(CardType: DWORD; AppWindow: HWND; ExtraInfo: Pointer=nil): BOOL;stdcall;
function DVBPLGClose(Reason: DWORD =0): BOOL;stdcall;
function DVBPLGCall(Code: DWORD =0; ExtraInfo: Pointer = nil): BOOL; stdcall;
function DVBPLGGetInfo(PLG_Info: pPLG_INFO): BOOL; stdcall;

exports DVBPLGInit    name 'DVBPLGInit';
exports DVBPLGClose   name 'DVBPLGClose';
exports DVBPLGCall    name 'DVBPLGCall';
exports DVBPLGGetInfo name 'DVBPLGGetInfo';

const
  MDAPI_VERSION_FOR_PLUGINS='MD-API Version 01.03 - 1.06';
  MDAPI_VERSION_FOR_PLUGINS9='MD-API Version 03.00 - 1.06';

// MD 8.2 Commands
  MDAPI_GET_TRANSPONDER       = $01020000;
  MDAPI_SET_TRANSPONDER       = $01020001;
  MDAPI_GET_PROGRAMM          = $01020010;
  MDAPI_SET_PROGRAMM          = $01020011;
  MDAPI_RESCAN_PROGRAMM       = $01020012;
  MDAPI_SAVE_PROGRAMM         = $01020013;
  MDAPI_GET_PROGRAMM_NUMMER   = $01020014;
  MDAPI_SET_PROGRAMM_NUMMER   = $01020015;
  MDAPI_START_FILTER          = $01020020;
  MDAPI_STOP_FILTER           = $01020021;
  MDAPI_SCAN_CURRENT_TP       = $01020030;
  MDAPI_SCAN_CURRENT_CAT      = $01020031;
  MDAPI_START_OSD             = $01020040;
  MDAPI_OSD_DRAWBLOCK         = $01020041;
  MDAPI_OSD_SETFONT           = $01020042;
  MDAPI_OSD_TEXT              = $01020043;
  MDAPI_SEND_OSD_KEY          = $01020044;
  MDAPI_STOP_OSD              = $01020049;
  MDAPI_DVB_COMMAND           = $01020060;
  MDAPI_GET_SIGNAL_INFO       = $01020061;
  MDAPI_GET_VERSION           = $01020100;

// ProgDVB commands
  PROGAPI_GET_SIGNAL_LEVEL    = $01022001;
  PROGAPI_SEND_DISEQC         = $01022002;
  PROGAPI_GET_CHANNEL_NAME    = $01022003;
  PROGAPI_SET_CHANNEL         = $01022004;
  PROGAPI_SET_TRANSPONDER     = $01022005;

// MD API 8.2
  MAX_CA_SYSTEMS=32;
  MAX_PID_IDS=32;
  STREAMTYPE_11172_VIDEO=1;
  STREAMTYPE_13818_VIDEO=2;
  STREAMTYPE_11172_AUDIO=3;
  STREAMTYPE_13818_AUDIO=4;
  STREAMTYPE_13818_PRIVATE=5;
  STREAMTYPE_13818_PES_PRIVATE=6;
  STREAMTYPE_13522_MHPEG=7;
  STREAMTYPE_13818_DSMCC=8;
  STREAMTYPE_ITU_222_1=9;
  STREAMTYPE_13818_A=10;
  STREAMTYPE_13818_B=11;
  STREAMTYPE_13818_C=12;
  STREAMTYPE_13818_D=13;
  STREAMTYPE_13818_AUX=14;

  MaxPluginCount=100;
  MaxFilterCount=100;
  MAX_BUFFER_SIZE=1024;
  TPCatCountMax=32;

// MD API 9.0
  MAX_CA_SYSTEMS90=48;

  MD_API_82=$0103;
  MD_API_90=$0300;

type
// MD API DataTypes
  TProgramNumberParam = record
    RealNumber: Integer;
    VirtNumber: Integer;
  end;
  PProgramNumberParam=^TProgramNumberParam;

  TStartFilterParam = record
    DLL_ID: Word;
    Filter_ID: Word;
    Pid: Word;
    Name: Array[0..32-1] of Byte;
    Irq_Call_Adresse: LongInt;
    Running_ID: Integer;
  end;
  PStartFilterParam=^TStartFilterParam;

  TTPCat = record
    TAG: Byte;
    DesLen: Byte;
    CA_ID: Word;
    EMM: Word;
    BufferLen: Word;
    Buffer: Array[0..64-1] of Byte;
  end;

  TTPCatio = record
    TPCatAnzahl: Integer;
    TPCat: Array[0..TPCatCountMax-1] of TTPCAT;
  end;
  PCAT=^TTPCatio;

  TPIDFilters = record
    FilterName: Array[0..5-1] of Char;
    FilterId: Byte;
    PID: Word;
  end;

  TCA_System82 = record
    CA_Typ:Word;
    ECM:Word;
    EMM:Word;
    Provider_Id:DWord;
  end;

  TProgramm82 = record
    Name: Array[00..29] of char;
    Anbieter:Array[00..29] of char;
    Land:Array[00..29] of char;
    freq: DWord;
    Typ: Byte;
    volt: Byte;
    afc: Byte;
    diseqc: Byte;
    srate:DWord;
    qam: Byte;
    fec: Byte;
    norm: Byte;
    tp_id: Word;
    Video_pid: Word;
    Audio_pid: Word;
    TeleText_pid:Word;
    PMT_pid:Word;
    PCR_pid:Word;
    ECM_PID:Word;
    SID_pid:Word;
    AC3_pid:Word;
    TVType: Byte;
    ServiceTyp: Byte;
    CA_ID: Byte;
    Temp_Audio:Word;
    Filteranzahl:Word;
    Filters: Array[00..(MAX_PID_IDS-1)] of TPIDFilters;
    CA_Anzahl:Word;
    CA_System82:array[0..(MAX_CA_SYSTEMS-1)] of TCA_System82;
    CA_Land: Array[0..5-1] of char;
    Merker:Byte;
    Link_TP:Word;
    Link_SID:Word;
    Dynamisch:Byte;
    Extern_Buffer:array[00..16-1] of char;
  end;
  PProgramm82=^TProgramm82;

  TProgramm90 = record
    Name: Array[00..29] of char;
    Anbieter:Array[00..29] of char;
    Land:Array[00..29] of char;
    freq: DWord;
    volt: Byte;
    diseqc: Byte;
//    alignment: Word;
    Typ: Byte; // alignment
    afc: Byte; // alignment
    srate: Word;
    qam: Word;
    fec: Word;
    norm: Byte;
    tp_id: Word;
    Video_pid: Word;
    Audio_pid: Word;
    TeleText_pid:Word;
    PMT_pid:Word;
    PCR_pid:Word;
    ECM_PID:Word;
    SID_pid:Word;
    AC3_pid:Word;
    TVType: Byte;
    ServiceTyp: Byte;
    CA_ID: Byte;
    Temp_Audio:Word;
    Filteranzahl:Word;
    Filters: Array[00..(MAX_PID_IDS-1)] of TPIDFilters;
    CA_Anzahl:Word;
    CA_System82:array[0..(MAX_CA_SYSTEMS90-1)] of TCA_System82;
    CA_Land: Array[0..5-1] of char;
    Merker:Byte;
    Link_TP:Word;
    Link_SID:Word;
    Dynamisch:Byte;
    Lists:array[0..12-1] of Word;  // Contains 12 Favorite-Sortings
    Current_i_id: Integer;
    Current_o_id: Integer;
    Next_i_id: Integer;
    Next_o_id: Integer;
    Extern_Buffer:array[00..15] of char;
  end;
  PProgramm90=^TProgramm90;

  TDVB_COMMAND = record
    Cmd_laenge: Word;
    Cmd_Buffer: Array[0..32-1] of Byte;
  end;
  PDVB_COMMAND=^TDVB_COMMAND;

  TSIGNAL_INFO = record
    SignalLevel: Cardinal;
    SignalError: Cardinal;
  end;
  PSIGNAL_INFO=^TSIGNAL_INFO;

  TTransponder = record
    ttk: Integer;
    power: Integer;
    diseqc: Integer;
    freq: DWord;
    volt: Integer;
    qam: Integer;
    srate: DWord;
    fec: Integer;
    sync: Integer;         //* sync from decoder */
    afc: Integer;          //* frequency offset in Hz */
    agc: Word;             //* gain */
    nest: Word;            //* noise estimation */
    vber: DWord;             //* viterbi bit error rate */
    ts_id: Word;
    sig: DWord;
    err: DWord;
  end;
  PTransponder=^TTransponder;

// MD API functions
  TMDAPIGetPluginName = procedure(Buf: PByteArray); cdecl;
  TMDAPIProcessPluginMenuCommandProc = procedure( MenuID:integer );cdecl;
  TMDAPIProcessChannelChange=procedure(prg: TProgramm82); cdecl;
  TMDAPIProcessChannelChange9=procedure(prg: TProgramm90); cdecl;
  TMDAPIInitPluginProc = procedure(MDInstance: LongWord; MDWnd: HWND; Log_Set: Bool; DLL_ID: Integer; HotKey: PChar;  Vers: PChar; var ReturnValue: Integer); cdecl;
  TMDAPIExitPluginProc = procedure(MDInstance: LongWord; MDWnd: HWND; Log_Set: Bool); cdecl;
  TMDAPIProcessHotKeyProc=procedure; cdecl;
  TMDAPIProcessOSDKeyProc=procedure(Osd_Key: Byte); cdecl;
  TMDAPIFilterProc=procedure( hFilter: DWORD; Len: Cardinal; Buf: PByteArray); cdecl;
  TMDAPIFilterCloseProc = procedure(MDInstance: LongWord); cdecl;
  TMDAPIProcessRecPlayProc = procedure(Mode:Integer); cdecl;
// ProgDVB types
  TProgAPIGetChannelName=record
    dwSize:DWORD; // in
    dwSID:DWORD; // in
    dwTID:DWORD; // in (if =0 so curretnt TID)
    dwSatPosition:DWORD ; // in (if =0 so current SatPosition)
    ChannelName:array [0..32-1] of byte; //out
    ProviderName:array [0..32-1] of byte; //out
    ChannelCaption: array [0..128-1] of byte; //out
  end;
  PProgAPIGetChannelName=^TProgAPIGetChannelName;

  TProgAPISignalState=record
    dwSize:DWORD;
    dwQuality:DWORD;
    dwLevel:DWORD;
  end;
  PProgAPISignalState=^TProgAPISignalState;

// Internal types
  TMDPluginWorkspace = class(TForm)
    MainMenu: TMainMenu;
    OptionsMenu: TMenuItem;
    PluginLog: TListBox;
    LogAllActions: TMenuItem;
    Acttions1: TMenuItem;
    LogLoadPlugin: TMenuItem;
    LogDetectPlugin: TMenuItem;
    LogCloseFilter: TMenuItem;
    LogFilterAction: TMenuItem;
    LogFilterCallback: TMenuItem;
    LogGetVersion: TMenuItem;
    LogGetProgramNumber: TMenuItem;
    LogSetProgramNumber: TMenuItem;
    LogGetProgram: TMenuItem;
    LogSetProgram: TMenuItem;
    LogSaveProgram: TMenuItem;
    LogScanCAT: TMenuItem;
    LogStartFilter: TMenuItem;
    LogStopFilter: TMenuItem;
    LogStartOSD: TMenuItem;
    LogOSDDrawBlock: TMenuItem;
    LogOSDSetFont: TMenuItem;
    LogOSDText: TMenuItem;
    LogSendOSDKey: TMenuItem;
    LogStopOSD: TMenuItem;
    LogDVBCommand: TMenuItem;
    LogGetSignalInfo: TMenuItem;
    LogPROGAPIGetSignalLevel: TMenuItem;
    MessageBoxOnError: TMenuItem;
    LogPROGAPIGetChannelName: TMenuItem;
    ChannelInfoOnlyFromCache: TMenuItem;
    GETCURRENTCATasSECTIONFilter1: TMenuItem;
    LogPROGAPISetChannelName: TMenuItem;
    LogGetTransponder: TMenuItem;
    LogSetTransponder: TMenuItem;
    LogRescanProgram: TMenuItem;
    View1: TMenuItem;
    Filters1: TMenuItem;
    Plugins1: TMenuItem;
    DelayAfterSettingFilter: TMenuItem;
    LogDVBCoreCOMMAND: TMenuItem;
    Packetsize: TMenuItem;
    P1: TMenuItem;
    P2: TMenuItem;
    P3: TMenuItem;
    P4: TMenuItem;
    P5: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PluginLogKeyPress(Sender: TObject; var Key: Char);
    procedure ViewAllFilters(Sender: TObject);
    procedure ViewAllPlugins(Sender: TObject);
    procedure BufferSizeChange(Sender: TObject);
  private
    { Private declarations }
    procedure GetCurrentChannelMDInfo(var prg:TProgramm82);
    procedure SetCurrentChannelMD(var Programm:PProgramm82);
    procedure GetCurrentChannelMD9Info(var prg:TProgramm90);
    procedure SetCurrentChannelMD9(var Programm:PProgramm90);
    function GetCurCAT(TPCatIO: PCAT): integer;
  public
    { Public declarations }
    ini:TIniFile;
    procedure InitMDPlugins;
    procedure ChangeChannel;
    procedure CloseFilters;
    procedure RecPlay(mode:Integer);
//    procedure LogAction(s:string;IsError:boolean=False);
    procedure WMCommand(var msg: TWMCommand); message WM_COMMAND;
    procedure WMUser(var msg : TMessage); message WM_USER;
    procedure WMEnable(var msg : TWMEnable); message WM_ENABLE;

  end;

  TMDPlugin=record
    Name:string;
    HotKey:char;
    MDAPI:Word;
    DLLID:Cardinal;
    PluginMenu:HMENU;
    //On_Start
    InitPlugin:                 TMDAPIInitPluginProc;
    //On_Send_Dll_ID_Name
    GetPluginName:              TMDAPIGetPluginName;
    //On_Menu_Select
    ProcessPluginMenuCommand:   TMDAPIProcessPluginMenuCommandProc;
    //On_Channel_Change
    ProcessChannelChange:       TMDAPIProcessChannelChange;
    ProcessChannelChange9:      TMDAPIProcessChannelChange9;
    //On_Exit
    ExitMDPlugin:               TMDAPIExitPluginProc;
    //On_Hot_Key
    ProcessHotKey:              TMDAPIProcessHotKeyProc;
    //On_Filter_Close
    ProcessFilterClose:         TMDAPIFilterCloseProc;
    //On_Rec_Play
    ProcessRecPlay:             TMDAPIProcessRecPlayProc;
    //On_Osd_Key
    ProcessOsdKey:              TMDAPIProcessOSDKeyProc;
    //On_VideoText_Stream
  end;

  TMDFilter=record
    Active:boolean;
    Handle:DWORD;
    Proc:TMDAPIFilterProc;
    DLL_ID: Word;
    Filter_ID: Word;
    Name:string;
    BufLen:Cardinal;
    Buffer: PByteArray;
    PID: Word;
  end;


var
  MDPluginWorkspace: TMDPluginWorkspace;
  Initialized:boolean=False;
  // Plugin API version
  MDAPIVersion:array[0..62] of byte;
  MDAPIVersion9:array[0..62] of byte;
  // You can change first bytes for different Firmware types
  SAA_COMMAND: array [00..13] of Byte=($10,4,5,0,0,0,0,0,0,0,0,0,0,0);
  CatIO: TTPCatIO;
  CATDone: boolean;
  CAT_CS: TRTLCriticalSection;
  CATEvent: DWORD;
  BUFFER_SIZE: WORD;
  NEW_BUFFER_SIZE: WORD;

  // Plugin information
  PluginCount:integer=0;
  MDPlugins: array [0..MaxPluginCount] of TMDPlugin;

  // Filter info for callback translation
  Filters:array [0..MaxFilterCount] of TMDFilter;

  // Saved channel parameters of MD-API but in memory (lack of DVBCore API)
  SavedTPInfo:TP_INFO;
  SavedPrgNum:integer=0;

  inited: boolean = false;

  CyrTable: array [0..255] of char = (
    ' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
    ' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
    ' ','!','"','#','$','%','&','''','(',')','*','+',',','-','.','/',
    '0','1','2','3','4','5','6','7','8','9',':',';','<','=','>','?',
    '@','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O',
    'P','Q','R','S','T','U','V','W','X','Y','Z','[','\',']','^','_',
    '`','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
    'p','q','r','s','t','u','v','w','x','y','z','{','|','}','~',' ',
    ' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
    ' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
    ' ','',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
    '','','','','','','','','','','','','','','','',
    '','','','','','','','','','','','','','','','',
    '','','','','','','','','','','','','','','','',
    '','','','','','','','','','','','','','','','',
    ' ','',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '
    );

    // Table of Translation Tables. Currently only Cyrillic table.
    // Please, send me other tables if you have it. or write it as described in
    // ETSI EN 300 468. I'm too lazy to write all tables :)

    TrnTable: array[0..31] of pointer =
    ( nil, @CyrTable, nil, nil, nil, nil, nil, nil,
      nil, nil,       nil, nil, nil, nil, nil, nil,
      nil, nil,       nil, nil, nil, nil, nil, nil,
      nil, nil,       nil, nil, nil, nil, nil, nil );

    MainHwnd: HWND;

implementation

{$R *.dfm}

procedure TMDPluginWorkspace.WMEnable(var msg : TWMEnable);
begin
    if (msg.Enabled=False) then EnableWindow(Handle,True);
    msg.Result:=0;
end;

function transl_str(str: PByteArray): string;
var
    i: integer;
    Transl: PChar;
begin

    result:='';
    Transl:=nil;

    for i:=1 to str[0]+1 do begin
        if str[i]=0 then break;

        if str[i]<$20 then begin
            Transl:=TrnTable[str[i]];
            continue;
        end;

        if (str[i] and $7F)>=$20 then begin
            if (Transl = nil) then result:=result+char(str[i])
                else result:=result+Transl[str[i]];
        end;
    end;
end;


procedure get_name(str, OutArray: PByteArray; Len: integer);
var
    i: integer;
begin
    for i:=0 to Len-1 do OutArray[i]:=0;
    StrCopy(pchar(OutArray),pchar(copy(transl_str(str),1,Len)));
end;

procedure LogAction(s:string;IsError:boolean=False);
const
    MAX_LOG_LINES=1000;
begin
    if NOT Initialized then Exit;

    if IsError and MDPluginWorkspace.MessageBoxOnError.Checked then
            ShowMessage(s);

    if MDPluginWorkspace.LogAllActions.Checked then //begin
        MDPluginWorkspace.PluginLog.Items.Insert(0,s);
// make horizontal scroll bar
{        MDPluginWorkspace.PluginLog.Canvas.Font:=MDPluginWorkspace.PluginLog.Font;
        if (MDPluginWorkspace.PluginLog.Canvas.TextWidth(s)+3>MDPluginWorkspace.PluginLog.ScrollWidth) then
            MDPluginWorkspace.PluginLog.ScrollWidth:=MDPluginWorkspace.PluginLog.Canvas.TextWidth(s)+3;
    end;}

    if MDPluginWorkspace.PluginLog.Items.Count>MAX_LOG_LINES then
        MDPluginWorkspace.PluginLog.Items.Delete(MAX_LOG_LINES);
end;

// Plugin Loading & Initialization
procedure TMDPluginWorkspace.InitMDPlugins;
var
    sr: TSearchRec;
    buf:array [0..511] of char;
    i:integer;
    Res:integer;
    Path_To_Plugins,s: string;
begin
    if NOT Initialized then begin
        Initialized:=True;

        StrPCopy(PChar(@MDAPIVersion[0]), MDAPI_VERSION_FOR_PLUGINS);
        StrPCopy(PChar(@MDAPIVersion9[0]), MDAPI_VERSION_FOR_PLUGINS9);
        for i:=0 to MaxFilterCount do Filters[i].Active:=False;
        for i:=0 to MaxPluginCount do MDPlugins[i].DLLID:=0;

        Path_To_Plugins:=extractfilepath(application.ExeName)+'MDPlugins\';

        if FindFirst(PATH_TO_PLUGINS+'*.dll', faAnyFile, sr) = 0 then begin
            repeat
                MDPlugins[PluginCount].DLLID:=LoadLibrary(PChar(PATH_TO_PLUGINS+sr.Name));
                if MDPlugins[PluginCount].DLLID<>0 then begin
                    MDPlugins[PluginCount].GetPluginName:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Send_Dll_ID_Name');
                    MDPlugins[PluginCount].InitPlugin:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Start');

                    if Assigned(MDPlugins[PluginCount].GetPluginName) and Assigned(MDPlugins[PluginCount].InitPlugin) then begin
                        MDPlugins[PluginCount].GetPluginName(@buf[0]);
                        MDPlugins[PluginCount].Name:=PChar(@buf[0]);
                        MDPlugins[PluginCount].ProcessPluginMenuCommand:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Menu_Select');
                        MDPlugins[PluginCount].ProcessChannelChange:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Channel_Change');
                        MDPlugins[PluginCount].ProcessFilterClose:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Filter_Close');
                        MDPlugins[PluginCount].ExitMDPlugin:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Exit');
                        MDPlugins[PluginCount].ProcessHotKey:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Hot_Key');
                        MDPlugins[PluginCount].ProcessOsdKey:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Osd_Key');
                        MDPlugins[PluginCount].ProcessRecPlay:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Rec_Play');

                        Res:=1;
                        s:=ini.ReadString('MD Plugins Cache', sr.Name+IntToStr(FileAge(PATH_TO_PLUGINS+sr.Name)),'0');
                        try
                            if ((StrToInt(s)=0) or ((StrToInt(s)<>MD_API_82) and (StrToInt(s)<>MD_API_90))) then begin
                                MDPlugins[PluginCount].MDAPI:=MD_API_82;
                                MDPlugins[PluginCount].InitPlugin(HInstance,Handle,true,PluginCount,@(MDPlugins[PluginCount].HotKey),@MDAPIVersion[0],Res);
                                if (Res=1) then begin
                                    MDPlugins[PluginCount].MDAPI:=MD_API_90;
                                    MDPlugins[PluginCount].InitPlugin(HInstance,Handle,true,PluginCount,@(MDPlugins[PluginCount].HotKey),@MDAPIVersion9[0],Res);
                                    MDPlugins[PluginCount].ProcessChannelChange9:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Channel_Change');
                                end;
                                ini.WriteString('MD Plugins Cache', sr.Name+IntToStr(FileAge(PATH_TO_PLUGINS+sr.Name)), IntToStr(MDPlugins[PluginCount].MDAPI));
                            end else if (StrToInt(s)=MD_API_82) then begin
                                MDPlugins[PluginCount].MDAPI:=MD_API_82;
                                MDPlugins[PluginCount].InitPlugin(HInstance,Handle,true,PluginCount,@(MDPlugins[PluginCount].HotKey),@MDAPIVersion[0],Res);
                            end else if (StrToInt(s)=MD_API_90) then begin
                                MDPlugins[PluginCount].MDAPI:=MD_API_90;
                                MDPlugins[PluginCount].InitPlugin(HInstance,Handle,true,PluginCount,@(MDPlugins[PluginCount].HotKey),@MDAPIVersion9[0],Res);
                                MDPlugins[PluginCount].ProcessChannelChange9:=GetProcAddress(MDPlugins[PluginCount].DLLID,'On_Channel_Change');
                            end;
                            if (Res=1) then begin
                                if LogLoadPlugin.Checked then LogAction('Not loaded: '+PChar(PATH_TO_PLUGINS+sr.Name),true);
                                if Assigned(MDPlugins[PluginCount].ExitMDPlugin) then
                                    MDPlugins[PluginCount].ExitMDPlugin(HInstance,Handle,true);
                                FreeLibrary(MDPlugins[PluginCount].DLLID);
                                MDPlugins[PluginCount].DLLID:=0;
                            end else begin
                                if LogLoadPlugin.Checked then LogAction('Loaded: '+PChar(PATH_TO_PLUGINS+sr.Name));
                                if LogDetectPlugin.Checked then
                                    if (MDPlugins[PluginCount].MDAPI=MD_API_82) then LogAction('Detected plugin: '+PChar(@buf[0])+' - MD-API 01.03')
                                    else LogAction('Detected plugin: '+PChar(@buf[0])+' - MD-API 03.00');
                                MDPlugins[PluginCount].PluginMenu:=LoadMenu(MDPlugins[PluginCount].DLLID,'EXTERN');
                                InsertMenu(Menu.Handle, $FFFFFFFF,(MF_BYPOSITION or MF_POPUP or MF_ENABLED),MDPlugins[PluginCount].PluginMenu,PChar(@buf[0]));
                                PluginCount:=PluginCount+1;
                            end;
                        except
                            if LogLoadPlugin.Checked then LogAction('Not loaded: '+PChar(PATH_TO_PLUGINS+sr.Name),true);
                            FreeLibrary(MDPlugins[PluginCount].DLLID);
                            MDPlugins[PluginCount].DLLID:=0;
                        end;
                    end else begin
                            if LogDetectPlugin.Checked then LogAction('Don''t MD plugin - '+sr.Name);
                            FreeLibrary(MDPlugins[PluginCount].DLLID);
                            MDPlugins[PluginCount].DLLID:=0;
                    end;
                end;
            until FindNext(sr) <> 0;
            FindClose(sr);
        end;
    end;
end;

// TP_INFO --> TProgramm82
procedure TMDPluginWorkspace.GetCurrentChannelMDInfo(var prg:TProgramm82);
var
    TP_Inf:TP_INFO;
    i,k,z,y:integer;
    tmp: word;
    CurA: word;
    TPCatIO: TTPCatio;
begin
    FillMemory(@prg, sizeof(prg), 0);
    if ChannelInfoOnlyFromCache.Checked then DVBParams(@TP_Inf,GET_ALLINFO or GET_ONLYFROMCACHE)
        else DVBParams(@TP_Inf,GET_ALLINFO or GET_TRYFROMCACHE);
    GetCurCat(@TPCatIO);

    if TP_Inf.wChCurrent=CURRENT_NONE then TP_Inf.wChCurrent:=0;

    get_name(@TP_Inf.Channel[TP_Inf.wChCurrent].Ch_Name, @prg.Name, 29);
    get_name(@TP_Inf.Channel[TP_Inf.wChCurrent].Prov_Name, @prg.Anbieter, 29);
  //prg.Land=??

    LogAction('DiSEqC len  : '+inttoHex(TP_Inf.bDiSEqLen,1));
    LogAction('DiSEqC tone : '+inttoHex(TP_Inf.bDiSEqToneburst,1));
    LogAction('DiSEqC data : '+inttoHex(TP_Inf.DiSEqData[0],1)+
    inttoHex(TP_Inf.DiSEqData[1],1)+
    inttoHex(TP_Inf.DiSEqData[2],1)+
    inttoHex(TP_Inf.DiSEqData[3],1)+
    inttoHex(TP_Inf.DiSEqData[4],1)+
    inttoHex(TP_Inf.DiSEqData[5],1)+
    inttoHex(TP_Inf.DiSEqData[6],1)+
    inttoHex(TP_Inf.DiSEqData[7],1)+
    inttoHex(TP_Inf.DiSEqData[8],1)+
    inttoHex(TP_Inf.DiSEqData[9],1)+
    inttoHex(TP_Inf.DiSEqData[10],1)+
    inttoHex(TP_Inf.DiSEqData[11],1)+
    inttoHex(TP_Inf.DiSEqData[12],1));

    if ((TP_Inf.bDiSEqLen<=32) and (TP_Inf.bDiSEqLen>0)) then
       prg.diseqc:=(TP_Inf.DiSEqData[TP_Inf.bDiSEqLen-1] and $0C) shr 2
    else if (TP_Inf.bDiSEqLen=255) then prg.diseqc:=(TP_Inf.DiSEqData[0] and $0C) shr 2
    else prg.diseqc:=0;
    prg.freq:=TP_Inf.dwFreq;
    prg.Typ:=byte('D'); // 'D' - Digital; 'A' - Analog;
    prg.volt:=byte(TP_Inf.dwLNB_Power);
    prg.srate:=word(TP_Inf.dwSymbRate div 1000);
    prg.fec:=TP_Inf.dwViterbi;
    prg.ServiceTyp:=byte(TP_Inf.Channel[Tp_Inf.wChCurrent].wType);
    prg.tp_id:=TP_Inf.wTSID;
    if (TP_Inf.dwLOF>4) then prg.qam:=2
    else prg.qam:=TP_Inf.dwLOF;

    prg.Video_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wVideo;

    CurA:=(TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCurrent and $FF);

    if CurA = CURRENT_NONE then CurA := 0;

    if ((TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.dwFlags[CurA] and DVB_AUDIO_TYPEMASK)=DVB_AUDIO_AC3) then
        prg.Audio_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPid[0]
    else
        prg.Audio_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPid[CurA];

    prg.AC3_pid:=$1FFF;
    prg.Filteranzahl:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCount;

    for i:=0 to TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCount-1 do begin
        if (i>MAX_PID_IDS-1) then break;
        k:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.dwFlags[i] and DVB_AUDIO_TYPEMASK;
        prg.Filters[i].PID:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPID[i];
        case k of
            DVB_AUDIO_AC3:
               begin
                 prg.AC3_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPID[i];
                 prg.Filters[i].FilterId:=STREAMTYPE_13818_PES_PRIVATE;
               end;
            DVB_AUDIO_ISO11172:
                 prg.Filters[i].FilterId:=STREAMTYPE_11172_AUDIO;
            DVB_AUDIO_ISO13818:
                 prg.Filters[i].FilterId:=STREAMTYPE_13818_AUDIO;
        end;
        if (i=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCurrent) then
            prg.Temp_Audio:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPID[i];
        CopyMemory(@prg.Filters[i].FilterName,PChar('aud'+IntToStr(i)),4);
    end;
    prg.PMT_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wPMT;
    prg.PCR_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wPCR;

    prg.TeleText_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wTeletext;

    if (TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent =  CURRENT_NONE)
        then TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent:=0;

    if (TP_Inf.EMM.wCurrent =  CURRENT_NONE)
        then TP_Inf.EMM.wCurrent:=0;

    tmp:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent];
    prg.ECM_PID:=tmp;
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent]:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[0];
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[0]:=tmp;

    tmp:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent];
    prg.CA_ID:=(tmp div 256);
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent]:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[0];
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[0]:=tmp;

    tmp:=TP_Inf.EMM.wPid[0];
    TP_Inf.EMM.wPid[0]:=TP_Inf.EMM.wPid[TP_Inf.EMM.wCurrent];
    TP_Inf.EMM.wPid[TP_Inf.EMM.wCurrent]:=tmp;

    tmp:=TP_Inf.EMM.wCA_Type[0];
    TP_Inf.EMM.wCA_Type[0]:=TP_Inf.EMM.wCA_Type[TP_Inf.EMM.wCurrent];
    TP_Inf.EMM.wCA_Type[TP_Inf.EMM.wCurrent]:=tmp;

    prg.SID_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wSID;

    prg.CA_Anzahl:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCount;
    if (prg.CA_Anzahl>MAX_CA_SYSTEMS-1) then prg.CA_Anzahl:=MAX_CA_SYSTEMS-1;

    for i:=0 to TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCount-1 do begin
        if (i>MAX_CA_SYSTEMS-1) then break;
        prg.CA_System82[i].CA_Typ:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[i];
        prg.CA_System82[i].ECM:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[i];
        for k:=0 to TPCatIO.TPCatAnzahl-1 do begin
            z:=11;
            for y:=0 to i do
                if (prg.CA_System82[i].CA_Typ=TPCatIO.TPCat[y].CA_ID) then
                     z:=z+5;
            if (prg.CA_System82[i].CA_Typ=TPCatIO.TPCat[k].CA_ID) then begin
                prg.CA_System82[i].EMM:=TPCatIO.TPCat[k].EMM;
                prg.CA_System82[i].Provider_Id:=(TPCatIO.TPCat[k].Buffer[z+0] shl 16)
                      or (TPCatIO.TPCat[k].Buffer[z+1] shl 8)
                      or (TPCatIO.TPCat[k].Buffer[z+2]);
                break;
            end;
        end;
    end;
    if (prg.ECM_PID=0) then prg.ECM_PID:=$1FFF;
    if (prg.AC3_pid=0) then prg.AC3_pid:=$1FFF;
    if (prg.TeleText_pid=0) then prg.TeleText_pid:=$1FFF;
    if (prg.Video_pid=0) then prg.Video_pid:=$1FFF;
    if (prg.Audio_pid=0) then prg.Audio_pid:=$1FFF;
    if (prg.PMT_pid=0) then prg.PMT_pid:=$1FFF;
    if (prg.PCR_pid=0) then prg.PCR_pid:=$1FFF;
    if (prg.SID_pid=0) then prg.SID_pid:=$1FFF;
    if (prg.Link_SID=0) then prg.Link_SID:=$FFFF;
    if (prg.Link_TP=0) then prg.Link_TP:=$FFFF;
    if (prg.CA_Anzahl=0) then begin
        prg.CA_System82[0].ECM:=$1FFF;
        prg.CA_System82[0].EMM:=$1FFF;
        prg.CA_System82[0].CA_Typ:=$FFFF;
        prg.CA_System82[0].Provider_Id:=$FFFFFFFF;
    end;
end;

// TProgramm82 --> TP_INFO
procedure TMDPluginWorkspace.SetCurrentChannelMD(var Programm:PProgramm82);
var
    TPInfo:TP_INFO;
    i,n,k,ch:integer;
begin
//  FillMemory(@TPInfo, sizeof(TPInfo), 0);
    if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
    else i:=GET_TRYFROMCACHE;
    DVBParams(@TPInfo, GET_ALLINFO or i);
    if TPInfo.wChCurrent=CURRENT_NONE then TPInfo.wChCurrent:=0;
    ch:=TPInfo.wChCurrent;
    TPInfo.Channel[ch].Ch_Name[0]:=Char(StrLen(@Programm.Name));
    TPInfo.Channel[ch].Prov_Name[0]:=Char(StrLen(@Programm.Anbieter));
    TPInfo.Channel[ch].Ch_Name[1]:=Chr(10);
    TPInfo.Channel[ch].Prov_Name[1]:=Chr(10);
    CopyMemory(@TPInfo.Channel[ch].Ch_Name[2],@Programm.Name,27);
    CopyMemory(@TPInfo.Channel[ch].Prov_Name[2],@Programm.Anbieter,27);
    if ((TPInfo.bDiSEqLen<=32) and (TPInfo.bDiSEqLen>0)) then
         TPInfo.DiSEqData[TPInfo.bDiSEqLen-1]:=TPInfo.DiSEqData[TPInfo.bDiSEqLen-1] or (Programm.diseqc shl 2)
    else if (TPInfo.bDiSEqLen=255) then
         TPInfo.DiSEqData[0]:=TPInfo.DiSEqData[0] or (Programm.diseqc shl 2);
    TPInfo.dwFreq:=Programm.freq;
    TPInfo.dwLNB_Power:=Programm.volt;
    TPInfo.dwSymbRate:=Programm.srate*1000;
    TPInfo.dwViterbi:=Programm.fec;
    TPInfo.Channel[ch].wType:=Programm.ServiceTyp;
    TPInfo.wTSID:=Programm.tp_id;
    TPInfo.Channel[ch].wVideo:=Programm.Video_pid;
    TPInfo.Channel[ch].Audio.wCount:=Programm.Filteranzahl;
    TPInfo.Channel[ch].Audio.wCurrent:=0;
    for i:=0 to Programm.Filteranzahl-1 do begin
        TPInfo.Channel[ch].Audio.wPid[i]:=Programm.Filters[i].PID;
        k:=TPInfo.Channel[ch].Audio.dwFlags[i] and ($FFFFFFFF xor DVB_AUDIO_TYPEMASK);
        case Programm.Filters[i].FilterId of
            STREAMTYPE_13818_PES_PRIVATE:
               TPInfo.Channel[ch].Audio.dwFlags[i]:=k or DVB_AUDIO_AC3;
            STREAMTYPE_11172_AUDIO:
               TPInfo.Channel[ch].Audio.dwFlags[i]:=k or DVB_AUDIO_ISO11172;
            STREAMTYPE_13818_AUDIO:
               TPInfo.Channel[ch].Audio.dwFlags[i]:=k or DVB_AUDIO_ISO13818;
        end;
        if (Programm.Filters[i].PID=Programm.Temp_Audio) then
            TPInfo.Channel[ch].Audio.wCurrent:=i;
    end;
    TPInfo.Channel[ch].wPMT:=Programm.PMT_pid;
    TPInfo.Channel[ch].wPCR:=Programm.PCR_pid;
    TPInfo.Channel[ch].wTeletext:=Programm.TeleText_pid;
    if (Programm.ECM_PID<>0) then TPInfo.Channel[ch].wScrambled:=16;
    TPInfo.Channel[ch].wSID:=Programm.SID_pid;
    TPInfo.Channel[ch].ECM.wCount:=0;
    TPInfo.EMM.wCount:=0;
    for i:=0 to Programm.CA_Anzahl-1 do begin
        if (Programm.CA_System82[i].ECM=0) then continue;
        TPInfo.Channel[ch].ECM.wCount:=TPInfo.Channel[ch].ECM.wCount+1;
        TPInfo.Channel[ch].ECM.wPid[i]:=Programm.CA_System82[i].ECM;
        TPInfo.Channel[ch].ECM.wCA_Type[i]:=Programm.CA_System82[i].CA_Typ;
        k:=0;
        for n:=0 to TPInfo.EMM.wCount do begin
             if (TPInfo.EMM.wPID[n]=Programm.CA_System82[i].EMM) then begin
                 k:=1;
                 break;
             end;
        end;
        if (k=0) then begin
            TPInfo.EMM.wPID[TPInfo.EMM.wCount]:=Programm.CA_System82[i].EMM;
            TPInfo.EMM.wCA_Type[TPInfo.EMM.wCount]:=Programm.CA_System82[i].CA_Typ;
            if (Programm.CA_System82[i].ECM=Programm.ECM_PID) then TPInfo.EMM.wCurrent:=TPInfo.EMM.wCount;
            TPInfo.EMM.wCount:=TPInfo.EMM.wCount+1;
        end;
        if (TPInfo.Channel[ch].ECM.wPid[i]=Programm.ECM_PID) then begin
            TPInfo.Channel[ch].wEMMCurrent:=Programm.CA_System82[i].EMM;
            TPInfo.wNetID:=Programm.CA_System82[i].Provider_Id;
            TPInfo.Channel[ch].ECM.wCurrent:=i;
        end;
    end;
    DVBParams(@TPInfo, SET_COPYALLDATA);
end;

// TP_INFO --> TProgramm90
procedure TMDPluginWorkspace.GetCurrentChannelMD9Info(var prg:TProgramm90);
var
    TP_Inf:TP_INFO;
    CAFound: boolean;
    i,k:integer;
    tmp: word;
    CurA: word;
begin
    FillMemory(@prg, sizeof(prg), 0);
    if ChannelInfoOnlyFromCache.Checked then DVBParams(@TP_Inf,GET_ALLINFO or GET_ONLYFROMCACHE)
        else DVBParams(@TP_Inf,GET_ALLINFO or GET_TRYFROMCACHE);

    if TP_Inf.wChCurrent=CURRENT_NONE then TP_Inf.wChCurrent:=0;

    get_name(@TP_Inf.Channel[TP_Inf.wChCurrent].Ch_Name, @prg.Name, 29);
    get_name(@TP_Inf.Channel[TP_Inf.wChCurrent].Prov_Name, @prg.Anbieter, 29);
  //prg.Land=??

    if ((TP_Inf.bDiSEqLen<=32) and (TP_Inf.bDiSEqLen>0)) then
       prg.diseqc:=(TP_Inf.DiSEqData[TP_Inf.bDiSEqLen-1] and $0C) shr 2
    else if (TP_Inf.bDiSEqLen=255) then prg.diseqc:=(TP_Inf.DiSEqData[0] and $0C) shr 2
    else prg.diseqc:=0;
    prg.freq:=TP_Inf.dwFreq;
    prg.volt:=byte(TP_Inf.dwLNB_Power);
    prg.srate:=word(TP_Inf.dwSymbRate div 1000);
    prg.fec:=TP_Inf.dwViterbi;
    prg.ServiceTyp:=byte(TP_Inf.Channel[Tp_Inf.wChCurrent].wType);
    prg.tp_id:=TP_Inf.wTSID;
    if (TP_Inf.dwLOF>4) then prg.qam:=2
    else prg.qam:=TP_Inf.dwLOF;

    prg.Video_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wVideo;

    CurA:=(TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCurrent and $FF);

    if CurA = CURRENT_NONE then CurA := 0;

    if ((TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.dwFlags[CurA] and DVB_AUDIO_TYPEMASK)=DVB_AUDIO_AC3) then
        prg.Audio_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPid[0]
    else
        prg.Audio_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPid[CurA];

    prg.AC3_pid:=$1FFF;
    prg.Filteranzahl:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCount;

    for i:=0 to TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCount-1 do begin
        if (i>MAX_PID_IDS-1) then break;
        k:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.dwFlags[i] and DVB_AUDIO_TYPEMASK;
        prg.Filters[i].PID:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPID[i];
        case k of
            DVB_AUDIO_AC3:
               begin
                 prg.AC3_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPID[i];
                 prg.Filters[i].FilterId:=STREAMTYPE_13818_PES_PRIVATE;
               end;
            DVB_AUDIO_ISO11172:
                 prg.Filters[i].FilterId:=STREAMTYPE_11172_AUDIO;
            DVB_AUDIO_ISO13818:
                 prg.Filters[i].FilterId:=STREAMTYPE_13818_AUDIO;
        end;         
        if (i=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wCurrent) then
            prg.Temp_Audio:=TP_Inf.Channel[Tp_Inf.wChCurrent].Audio.wPID[i];
        CopyMemory(@prg.Filters[i].FilterName,PChar('aud'+IntToStr(i)),4);
    end;
    prg.PMT_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wPMT;
    prg.PCR_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wPCR;

    prg.TeleText_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wTeletext;

    if (TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent =  CURRENT_NONE)
        then TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent:=0;

    if (TP_Inf.EMM.wCurrent =  CURRENT_NONE)
        then TP_Inf.EMM.wCurrent:=0;

    tmp:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent];
    prg.ECM_PID:=tmp;
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent]:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[0];
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[0]:=tmp;

    tmp:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent];
    prg.CA_ID:=(tmp div 256);
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCurrent]:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[0];
    TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[0]:=tmp;

    tmp:=TP_Inf.EMM.wPid[0];
    TP_Inf.EMM.wPid[0]:=TP_Inf.EMM.wPid[TP_Inf.EMM.wCurrent];
    TP_Inf.EMM.wPid[TP_Inf.EMM.wCurrent]:=tmp;

    tmp:=TP_Inf.EMM.wCA_Type[0];
    TP_Inf.EMM.wCA_Type[0]:=TP_Inf.EMM.wCA_Type[TP_Inf.EMM.wCurrent];
    TP_Inf.EMM.wCA_Type[TP_Inf.EMM.wCurrent]:=tmp;

    prg.SID_pid:=TP_Inf.Channel[Tp_Inf.wChCurrent].wSID;

    prg.CA_Anzahl:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCount;
    if (prg.CA_Anzahl>MAX_CA_SYSTEMS90-1) then prg.CA_Anzahl:=MAX_CA_SYSTEMS90-1;

    for i:=0 to TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCount-1 do begin
        if (i>MAX_CA_SYSTEMS90-1) then break;
        prg.CA_System82[i].CA_Typ:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wCA_Type[i];
        prg.CA_System82[i].ECM:=TP_Inf.Channel[Tp_Inf.wChCurrent].ECM.wPid[i];
        prg.CA_System82[i].Provider_Id:=TP_Inf.wNetID;
    end;

    for i:=0 to TP_Inf.EMM.wCount-1 do begin
        CAFound:=false;
        for k:=0 to prg.CA_Anzahl-1 do begin
            if (prg.CA_System82[k].CA_Typ = TP_Inf.EMM.wCA_Type[i]) and (prg.CA_System82[k].EMM =0) then begin
                prg.CA_System82[k].EMM:=TP_Inf.EMM.wPID[i];
                CAFound:=true;
            end;
        end;
        if not CAFound then begin
            prg.CA_System82[prg.CA_Anzahl].CA_Typ := TP_Inf.EMM.wCA_Type[i];
            prg.CA_System82[prg.CA_Anzahl].EMM:=TP_Inf.EMM.wPID[i];
            prg.CA_System82[prg.CA_Anzahl].Provider_Id:=TP_Inf.wNetID;
            inc(prg.CA_Anzahl);
        end;
    end;
    if (prg.ECM_PID=0) then prg.ECM_PID:=$1FFF;
    if (prg.AC3_pid=0) then prg.AC3_pid:=$1FFF;
    if (prg.TeleText_pid=0) then prg.TeleText_pid:=$1FFF;
    if (prg.Video_pid=0) then prg.Video_pid:=$1FFF;
    if (prg.Audio_pid=0) then prg.Audio_pid:=$1FFF;
    if (prg.PMT_pid=0) then prg.PMT_pid:=$1FFF;
    if (prg.PCR_pid=0) then prg.PCR_pid:=$1FFF;
    if (prg.SID_pid=0) then prg.SID_pid:=$1FFF;
    if (prg.Link_SID=0) then prg.Link_SID:=$FFFF;
    if (prg.Link_TP=0) then prg.Link_TP:=$FFFF;
    if (prg.CA_Anzahl=0) then begin
        prg.CA_System82[0].ECM:=$1FFF;
        prg.CA_System82[0].EMM:=$1FFF;
        prg.CA_System82[0].CA_Typ:=$FFFF;
        prg.CA_System82[0].Provider_Id:=$FFFFFFFF;
    end;
end;

// TProgramm90 --> TP_INFO
procedure TMDPluginWorkspace.SetCurrentChannelMD9(var Programm:PProgramm90);
var
    TPInfo:TP_INFO;
    i,n,k,ch:integer;
begin
//  FillMemory(@TPInfo, sizeof(TPInfo), 0);
    if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
    else i:=GET_TRYFROMCACHE;
    DVBParams(@TPInfo, GET_ALLINFO or i);
    if TPInfo.wChCurrent=CURRENT_NONE then TPInfo.wChCurrent:=0;
    ch:=TPInfo.wChCurrent;
    TPInfo.Channel[ch].Ch_Name[0]:=Char(StrLen(@Programm.Name));
    TPInfo.Channel[ch].Prov_Name[0]:=Char(StrLen(@Programm.Anbieter));
    TPInfo.Channel[ch].Ch_Name[1]:=Chr(10);
    TPInfo.Channel[ch].Prov_Name[1]:=Chr(10);
    CopyMemory(@TPInfo.Channel[ch].Ch_Name[2],@Programm.Name,27);
    CopyMemory(@TPInfo.Channel[ch].Prov_Name[2],@Programm.Anbieter,27);
    if ((TPInfo.bDiSEqLen<=32) and (TPInfo.bDiSEqLen>0)) then
         TPInfo.DiSEqData[TPInfo.bDiSEqLen-1]:=TPInfo.DiSEqData[TPInfo.bDiSEqLen-1] or (Programm.diseqc shl 2)
    else if (TPInfo.bDiSEqLen=255) then
         TPInfo.DiSEqData[0]:=TPInfo.DiSEqData[0] or (Programm.diseqc shl 2);
    TPInfo.dwFreq:=Programm.freq;
    TPInfo.dwLNB_Power:=Programm.volt;
    TPInfo.dwSymbRate:=Programm.srate*1000;
    TPInfo.dwViterbi:=Programm.fec;
    TPInfo.Channel[ch].wType:=Programm.ServiceTyp;
    TPInfo.wTSID:=Programm.tp_id;
    TPInfo.Channel[ch].wVideo:=Programm.Video_pid;
    TPInfo.Channel[ch].Audio.wCount:=Programm.Filteranzahl;
    TPInfo.Channel[ch].Audio.wCurrent:=0;
    for i:=0 to Programm.Filteranzahl-1 do begin
        TPInfo.Channel[ch].Audio.wPid[i]:=Programm.Filters[i].PID;
        k:=TPInfo.Channel[ch].Audio.dwFlags[i] and ($FFFFFFFF xor DVB_AUDIO_TYPEMASK);
        case Programm.Filters[i].FilterId of
            STREAMTYPE_13818_PES_PRIVATE:
               TPInfo.Channel[ch].Audio.dwFlags[i]:=k or DVB_AUDIO_AC3;
            STREAMTYPE_11172_AUDIO:
               TPInfo.Channel[ch].Audio.dwFlags[i]:=k or DVB_AUDIO_ISO11172;
            STREAMTYPE_13818_AUDIO:
               TPInfo.Channel[ch].Audio.dwFlags[i]:=k or DVB_AUDIO_ISO13818;
        end;
        if (Programm.Filters[i].PID=Programm.Temp_Audio) then
            TPInfo.Channel[ch].Audio.wCurrent:=i;
    end;
    TPInfo.Channel[ch].wPMT:=Programm.PMT_pid;
    TPInfo.Channel[ch].wPCR:=Programm.PCR_pid;
    TPInfo.Channel[ch].wTeletext:=Programm.TeleText_pid;
    if (Programm.ECM_PID<>0) then TPInfo.Channel[ch].wScrambled:=16;
    TPInfo.Channel[ch].wSID:=Programm.SID_pid;
    TPInfo.Channel[ch].ECM.wCount:=0;
    TPInfo.EMM.wCount:=0;
    for i:=0 to Programm.CA_Anzahl-1 do begin
        if (Programm.CA_System82[i].ECM=0) then continue;
        TPInfo.Channel[ch].ECM.wCount:=TPInfo.Channel[ch].ECM.wCount+1;
        TPInfo.Channel[ch].ECM.wPid[i]:=Programm.CA_System82[i].ECM;
        TPInfo.Channel[ch].ECM.wCA_Type[i]:=Programm.CA_System82[i].CA_Typ;
        k:=0;
        for n:=0 to TPInfo.EMM.wCount do begin
             if (TPInfo.EMM.wPID[n]=Programm.CA_System82[i].EMM) then begin
                 k:=1;
                 break;
             end;
        end;
        if (k=0) then begin
            TPInfo.EMM.wPID[TPInfo.EMM.wCount]:=Programm.CA_System82[i].EMM;
            TPInfo.EMM.wCA_Type[TPInfo.EMM.wCount]:=Programm.CA_System82[i].CA_Typ;
            if (Programm.CA_System82[i].ECM=Programm.ECM_PID) then TPInfo.EMM.wCurrent:=TPInfo.EMM.wCount;
            TPInfo.EMM.wCount:=TPInfo.EMM.wCount+1;
        end;
        if (TPInfo.Channel[ch].ECM.wPid[i]=Programm.ECM_PID) then begin
            TPInfo.Channel[ch].wEMMCurrent:=Programm.CA_System82[i].EMM;
            TPInfo.wNetID:=Programm.CA_System82[i].Provider_Id;
            TPInfo.Channel[ch].ECM.wCurrent:=i;
        end;
    end;
    DVBParams(@TPInfo, SET_COPYALLDATA);
end;

procedure TMDPluginWorkspace.RecPlay(mode:Integer);
var i:integer;
begin
    for i:=0 to PluginCount-1 do
        if (MDPlugins[i].DLLID<>0) and Assigned(MDPlugins[i].ProcessRecPlay) then MDPlugins[i].ProcessRecPlay(mode);
end;

procedure TMDPluginWorkspace.ChangeChannel;
var prg:TProgramm82;
    prg9:TProgramm90;
    i:integer;
begin
    GetCurrentChannelMDInfo(prg);
    GetCurrentChannelMD9Info(prg9);
    for i:=0 to PluginCount-1 do
        if (MDPlugins[i].DLLID<>0) and (Assigned(MDPlugins[i].ProcessChannelChange) or Assigned(MDPlugins[i].ProcessChannelChange9)) then begin
            if MDPluginWorkspace.LogDVBCoreCOMMAND.Checked then
                 LogAction('CHANNEL CHANGE: for Plugin'+IntToStr(i+1));
            if (MDPlugins[i].MDAPI=MD_API_82) then MDPlugins[i].ProcessChannelChange(prg)
            else MDPlugins[i].ProcessChannelChange9(prg9);
        end;
end;

procedure TMDPluginWorkspace.CloseFilters;
var i:integer;
begin
    for i:=0 to MaxFilterCount do
        if Filters[i].Active then begin
            if (MDPlugins[Filters[i].DLL_ID].DLLID<>0) and Assigned(MDPlugins[Filters[i].DLL_ID].ProcessFilterClose) then begin
                if MDPluginWorkspace.LogDVBCoreCOMMAND.Checked then
                    LogAction('ON FILTER CLOSE: to Plugin'+IntToStr(Filters[i].DLL_ID+1)+' for Filter'+IntToStr(i+1));
                MDPlugins[Filters[i].DLL_ID].ProcessFilterClose(Filters[i].Filter_ID);
            end;
            if LogCloseFilter.Checked then LogAction('Closed Filter: '+IntToStr(i+1)+'['+IntToHex(Filters[i].Handle,8)+']');
            Filters[i].Active:=False;
            DVBDelFilter(Filters[i].Handle);
            if (Filters[i].Buffer<>nil) then
                LocalFree(Cardinal(Filters[i].Buffer));
            Filters[i].Buffer:=nil;
        end;
end;

procedure TMDPluginWorkspace.WMCommand(var msg: TWMCommand);
var i:integer;
begin
    for i:=0 to PluginCount-1 do
        if (MDPlugins[i].DLLID<>0) and Assigned(MDPlugins[i].ProcessPluginMenuCommand) then MDPlugins[i].ProcessPluginMenuCommand(msg.ItemID);
    inherited;
end;

procedure DVBCore2MDFilterThunk(Buf: PByteArray;  Len: Cardinal;  hFilter: DWORD); stdcall;
var
    i:Integer;
    tempbuf: Array[0..1024-1] of Byte;
    tempbuflen:Cardinal;
begin
    if Not Initialized then Exit;

    for i:=0 to MaxFilterCount do
        if (Filters[i].Active and (Filters[i].Handle=hFilter)) then begin
//             Filters[i].Proc(Filters[i].Filter_ID,Len,Buf);
            if (Filters[i].Buffer<>nil) then
                if (Filters[i].BufLen+Len>=NEW_BUFFER_SIZE-Len) then begin
                    tempbuflen:=Filters[i].BufLen+Len;
                    CopyMemory(@Filters[i].Buffer[Filters[i].BufLen],Buf,Len);
                    CopyMemory(@tempbuf,Filters[i].Buffer,tempbuflen);
                    if MDPluginWorkspace.LogFilterCallback.Checked then
                        LogAction('Filter Callback : '+IntToStr(i+1)+'['+inttoHex(Filters[i].Handle,8)+'], Buffer Length='+IntToStr(tempbuflen));
                    Filters[i].Proc(Filters[i].Filter_ID,tempbuflen,@tempbuf);
                    Filters[i].BufLen:=0;
                end else begin
                    CopyMemory(@Filters[i].Buffer[Filters[i].BufLen],Buf,Len);
                    Filters[i].BufLen:=Filters[i].BufLen+Len;
                end;
             break;
       end;
end;

procedure CATSecFlt(Buf: PByteArray;  Len: Cardinal;  hFilter: DWORD); stdcall;
var
    SecEnd,i: Cardinal;
begin
    SecEnd:=(((Buf[1]*256) + Buf[2]) and $FFF)+3;
    if SecEnd>Len then exit;
    i:=8;
    while(i<SecEnd) do begin
        if Buf[i]=9 then begin
            CatIO.TPCat[CatIO.TPCatAnzahl].TAG:=Buf[i];
            CatIO.TPCat[CatIO.TPCatAnzahl].DesLen:=Buf[i+1];
            CatIO.TPCat[CatIO.TPCatAnzahl].CA_ID:=((Buf[i+2]*256) + Buf[i+3]);
            CatIO.TPCat[CatIO.TPCatAnzahl].EMM:=((Buf[i+4]*256) + Buf[i+5]) and $1FFF;
            CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen:=Buf[i+1]+2;
            if CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen>64 then CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen:=64;
            CopyMemory(@CatIO.TPCat[CatIO.TPCatAnzahl].Buffer, @Buf[i], CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen);
            inc(CatIO.TPCatAnzahl);
            if (CatIO.TPCatAnzahl>TPCatCountMax-1) then break;
        end;
        i:=i+Buf[i+1]+2;
    end;
    if (CatIO.TPCatAnzahl>0) then SetEvent(CATEvent);
end;


procedure CATStrmFlt(Buf: PByteArray;  Len: Cardinal;  hFilter: DWORD); stdcall;
var
    SecEnd,i,k: Cardinal;
begin
    k:=0;
    while (k<Len) do begin
        i:=k;
        if (Buf[i]=0) and (Buf[i+1]=1) then begin
            SecEnd:=(((Buf[i+2]*256) + Buf[i+3]) and $FFF)+i+3;
            if SecEnd>Len then exit;
            i:=i+9;
            while(i<SecEnd) do begin
                if Buf[i]=9 then begin
                    CatIO.TPCat[CatIO.TPCatAnzahl].TAG:=Buf[i];
                    CatIO.TPCat[CatIO.TPCatAnzahl].DesLen:=Buf[i+1];
                    CatIO.TPCat[CatIO.TPCatAnzahl].CA_ID:=((Buf[i+2]*256) + Buf[i+3]);
                    CatIO.TPCat[CatIO.TPCatAnzahl].EMM:=((Buf[i+4]*256) + Buf[i+5]) and $1FFF;
                    CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen:=Buf[i+1]+2;
                    if CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen>64 then CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen:=64;
                    CopyMemory(@CatIO.TPCat[CatIO.TPCatAnzahl].Buffer, @Buf[i], CatIO.TPCat[CatIO.TPCatAnzahl].BufferLen);
                    inc(CatIO.TPCatAnzahl);
                    if (CatIO.TPCatAnzahl>TPCatCountMax-1) then break;
                end;
                i:=i+Buf[i+1]+2;
            end;
            if (CatIO.TPCatAnzahl>0) then break;
        end;
        k:=k+184;
    end;
    if (CatIO.TPCatAnzahl>0) then SetEvent(CATEvent);

end;


function TMDPluginWorkspace.GetCurCAT(TPCatIO: PCAT): integer;
var
    i, hFilter: Cardinal;
    data: array[0..5] of byte;
    mask: array[0..5] of byte;
begin
    data[0]:=1;
    mask[0]:=$FF;
    EnterCriticalSection(CAT_CS);
    ResetEvent(CATEvent);
    CatIO.TPCatAnzahl:=0;
    result:=0;
    if MDPluginWorkspace.GETCURRENTCATasSECTIONFilter1.Checked
        then hFilter:=DVBAddFilter(1,@CatSecFlt, FT_SECTION_FILTER, @data[0], @mask[0],1)
        else hFilter:=DVBAddFilter(1,@CatStrmFlt);
    if hFilter>0 then begin
        i:=WaitForSingleObject(CATEvent, 3000);
        DVBDelFilter(hFilter);
        if (i=WAIT_OBJECT_0) and (CatIO.TPCatAnzahl>0) then begin
            CopyMemory(TPCatIO, @CatIO, sizeof(TTPCatIO));
            Result:=1;
        end;
    end;
    LeaveCriticalSection(CAT_CS);
end;

procedure TMDPluginWorkspace.WMUser(var msg : TMessage);
var
    TPInfo:TP_INFO;
    i,n,k:integer;
    StartFilterParam:PStartFilterParam;
    DVBTunInfo:DVB_TUNER_INFO;
    MDSignalInfo:PSIGNAL_INFO;
    ProgAPISignalState:PProgAPISignalState;
    ProgAPIGetChannelName:PProgAPIGetChannelName;
    Transponder:PTransponder;
//    DSDisplay:DS_DISPLAY;
begin
    if Not Initialized then Exit;

    case msg.WParam of
        MDAPI_GET_VERSION:
            begin
                if LogGetVersion.Checked then LogAction('GET_VERSION');
                StrPCopy(PChar(msg.LParam), MDAPI_VERSION_FOR_PLUGINS9);
                msg.Result:=1;
                exit;
            end;

        MDAPI_GET_PROGRAMM_NUMMER:
            begin
                if LogGetProgramNumber.Checked then LogAction('GET_PROGRAMM_NUMMER');
                PProgramNumberParam(msg.LParam)^.RealNumber:=SavedPrgNum;
                PProgramNumberParam(msg.LParam)^.VirtNumber:=SavedPrgNum;
                msg.Result:=1;
                exit;
            end;

        MDAPI_SET_PROGRAMM_NUMMER:
            begin
                if ((msg.LParam=SavedPrgNum) and (SavedTPInfo.wDataValid<>0)) then begin
                    if LogSetProgramNumber.Checked then LogAction('SET_PROGRAMM_NUMMER - OK');
//                    not worked, result structure is empty, but need or not? (Scrib)
//                    DVBDSDisplay(@DSDisplay,DVB_DSDISP_GETSTATE);
//                    if (DSDisplay.State=DVB_DSDISP_PLAY) then
                        i:=SET_STARTCAPTURE;
//                    else i:=SET_STOPCAPTURE;
                    DVBParams(@SavedTPInfo, SET_COPYALLDATA or SET_TUNER or SET_PIDS or i);
                    msg.Result:=1;
                end else
                    if LogSetProgramNumber.Checked then LogAction('SET_PROGRAMM_NUMMER - FAIL');
                    msg.Result:=0;
                exit;
            end;

        MDAPI_GET_TRANSPONDER:
            begin
                if LogGetTransponder.Checked then LogAction('GET_TRANSPONDER');
                Transponder:=PTransponder(msg.LParam);
                FillMemory(@Transponder^, sizeof(Transponder), 0);
                if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
                else i:=GET_TRYFROMCACHE;
                DVBParams(@TPInfo, GET_ALLINFO or i);
                Transponder^.fec:=TPInfo.dwViterbi;
                if ((TPInfo.bDiSEqLen<=32) and (TPInfo.bDiSEqLen>0)) then
                    Transponder^.diseqc:=(TPInfo.DiSEqData[TPInfo.bDiSEqLen-1] and $0C) shr 2
                else  if (TPInfo.bDiSEqLen=255) then Transponder^.diseqc:=(TPInfo.DiSEqData[0] and $0C) shr 2
                else Transponder^.diseqc:=0;
                Transponder^.freq:=TPInfo.dwFreq;
                if (TPInfo.dwLOF>4) then Transponder^.qam:=2
                else Transponder^.qam:=TPInfo.dwLOF;
                Transponder^.srate:=word(TPInfo.dwSymbRate div 1000);
                Transponder^.volt:=byte(TPInfo.dwLNB_Power);
                Transponder^.sync:=3;
                Transponder^.ts_id:=TPInfo.wTSID;
                msg.Result:=1;
                exit;
            end;

        MDAPI_SET_TRANSPONDER:
            begin
                if LogSetTransponder.Checked then LogAction('SET_TRANSPONDER');
                Transponder:=PTransponder(msg.LParam);
                if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
                else i:=GET_TRYFROMCACHE;
                DVBParams(@TPInfo, GET_ALLINFO or i);
                TPInfo.dwViterbi:=Transponder.fec;
                if ((TPInfo.bDiSEqLen<=32) and (TPInfo.bDiSEqLen>0)) then
                    TPInfo.DiSEqData[TPInfo.bDiSEqLen-1]:=TPInfo.DiSEqData[TPInfo.bDiSEqLen-1] or (Transponder.diseqc shl 2)
                else if (TPInfo.bDiSEqLen=255) then
                    TPInfo.DiSEqData[0]:=TPInfo.DiSEqData[0] or (Transponder.diseqc shl 2);
                TPInfo.dwFreq:=Transponder.freq;
                TPInfo.dwSymbRate:=Transponder.srate*1000;
                TPInfo.dwLNB_Power:=Transponder.volt;
                TPInfo.wTSID:=Transponder.ts_id;
                DVBParams(@TPInfo, SET_TUNER);
                msg.Result:=1;
                exit;
            end;

        MDAPI_GET_PROGRAMM:
            begin
                if LogGetProgram.Checked then LogAction('GET_PROGRAMM');
                // no possibility to determine what plugin send this message
                GetCurrentChannelMDInfo(PProgramm82(msg.LParam)^);
                msg.Result:=1;
                exit;
            end;

        MDAPI_RESCAN_PROGRAMM,
        MDAPI_SET_PROGRAMM:
            begin
                if (msg.WParam=MDAPI_RESCAN_PROGRAMM) then
                    if LogRescanProgram.Checked then LogAction('RESCAN_PROGRAMM')
                else
                    if LogSetProgram.Checked then LogAction('SET_PROGRAMM');
                if (PProgramm90(msg.LParam).diseqc>3) then i:=MD_API_82 //MD-API 01.03 (PProgramm82(msg.LParam).volt 13||18)
                else i:=MD_API_90; //MD-API 03.00
                if (i=MD_API_90) then SetCurrentChannelMD9(PProgramm90(msg.LParam))
                else SetCurrentChannelMD(PProgramm82(msg.LParam));
                if (msg.WParam=MDAPI_RESCAN_PROGRAMM) then
                    if (i=MD_API_90) then GetCurrentChannelMD9Info(PProgramm90(msg.LParam)^)
                    else GetCurrentChannelMDInfo(PProgramm82(msg.LParam)^);
                msg.Result:=1;
                exit;
            end;

        MDAPI_SAVE_PROGRAMM:
            begin
                if LogSaveProgram.Checked then LogAction('SAVE_PROGRAMM');
                SavedPrgNum:=SavedPrgNum+1;
                if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
                else i:=GET_TRYFROMCACHE;
                DVBParams(@SavedTPInfo, GET_ALLINFO or i);
                // impossible, DVBCore don't managed channel list (Scrib)
                // but Ok - for compatibility
                msg.Result:=1;
                exit;
            end;

        MDAPI_SCAN_CURRENT_CAT:
            begin
                msg.Result:=GetCurCat(PCAT(msg.LParam));
                if LogScanCAT.Checked then
                   if (msg.Result=1) then LogAction('SCAN_CURRENT_CAT - OK')
                   else LogAction('SCAN_CURRENT_CAT - FAIL',true);
                exit;
            end;

        MDAPI_START_FILTER:
            begin
                i:=0;
                while ((i<=MaxFilterCount) and (Filters[i].Active)) do i:=i+1;
                if i>MaxFilterCount then begin
                    LogAction('Maximum Filter Count Exceeded!!!',True);
                    msg.Result:=0;
                end else begin
                    StartFilterParam:=PStartFilterParam(msg.LParam);
                    if ((StartFilterParam^.Pid<>0) and (StartFilterParam^.Pid<>$1FFF)) then begin
                      if LogStartFilter.Checked then LogAction('START_FILTER: '+PChar(@(StartFilterParam^.Name[0])));
                      for n:=0 to MaxFilterCount do if (Filters[n].Active and (CompareStr(Filters[n].Name,PChar(@(StartFilterParam^.Name[0])))=0) and (Filters[n].PID=StartFilterParam^.Pid and $1FFF)) then begin
                          if LogStartFilter.Checked then begin
                              LogAction('Add filter without delete!!');
                              LogAction('Force delete previous declaration: '+inttohex(Filters[n].Handle,8));
                          end;
                          Filters[n].Active:=False;
                          DVBDelFilter(Filters[n].Handle);
                          if (Filters[n].Buffer<>nil) then
                              LocalFree(Cardinal(Filters[i].Buffer));
                          break;
                      end;
                      Filters[i].BufLen:=0;
                      BUFFER_SIZE:=NEW_BUFFER_SIZE;
                      Filters[i].Buffer:=Pointer(LocalAlloc(LMEM_FIXED,MAX_BUFFER_SIZE));
                      if (Filters[i].Buffer<>nil) then begin
                         Filters[i].PID:=StartFilterParam^.Pid and $1FFF;
                         Filters[i].Filter_ID:=StartFilterParam^.Filter_ID;
                         Filters[i].Proc:=TMDAPIFilterProc(StartFilterParam^.Irq_Call_Adresse);
                         Filters[i].Active:=True;
                         Filters[i].Handle:=DVBAddFilter(Filters[i].PID,@DVBCore2MDFilterThunk,FT_PIPING_FILTER);
                         Filters[i].Name:=PChar(@(StartFilterParam^.Name[0]));
                         Filters[i].DLL_ID:=StartFilterParam^.DLL_ID;
                         StartFilterParam^.Running_ID:=(Filters[i].PID shl 8) + (i+1);
                         if DelayAfterSettingFilter.Checked then begin
                             Sleep(3);Sleep(3);Sleep(3);Sleep(3);Sleep(3);
                             Sleep(3);Sleep(3);Sleep(3);Sleep(3);Sleep(3);
                             Sleep(3);Sleep(3);Sleep(3);Sleep(3);Sleep(3);
                         end;
                         if LogStartFilter.Checked then
                             LogAction('Added Filter : '+IntToStr(i+1)+'['+inttoHex(Filters[i].Handle,8)+'] ('+IntToHex(StartFilterParam^.Pid,4)+')');
                         msg.Result:=1;
                      end else begin
                         if LogStartFilter.Checked then LogAction('Can not allocate memory for buffer!!!',true);
                         msg.Result:=0;
                      end;
                    end else begin
                      if LogStartFilter.Checked then LogAction('START_FILTER: '+PChar(@(StartFilterParam^.Name[0]))+' - zero PID skipped');
                      msg.Result:=0;
                    end;
                end;
                exit;
            end;

        MDAPI_STOP_FILTER:
            begin
                i:=integer(msg.LParam);
                k:=i shr 8;          // PID
                i:=(i and $FF)-1;    // index
                if LogStopFilter.Checked then LogAction('MDAPI_STOP_FILTER: PID '+IntToHex(k,4));
                if (Filters[i].Active and (Filters[i].PID=k)) then begin
                    Filters[i].Active:=False;
                    DVBDelFilter(Filters[i].Handle);
                    if (Filters[i].Buffer<>nil) then
                        LocalFree(Cardinal(Filters[i].Buffer));
                    Filters[i].Buffer:=nil;
                    if LogStopFilter.Checked then LogAction('Stoped Filter : '+IntToStr(i+1)+'['+IntToHex(Filters[i].Handle,8)+']');
                    msg.Result:=1;
                end else msg.Result:=0;
                exit;
            end;

        MDAPI_GET_SIGNAL_INFO:
            begin
                if LogGetSignalInfo.Checked then LogAction('GET_SIGNAL_INFO');
                DVBGetTunerInfo(@DVBTunInfo);
                MDSignalInfo:=PSIGNAL_INFO(msg.LParam);
                MDSignalInfo^.SignalLevel:=(DVBTunInfo.Level*255) div 100;
                MDSignalInfo^.SignalError:=DVBTunInfo.SNR_SQE;
                msg.Result:=1;
                exit;
            end;

        MDAPI_START_OSD:
            if LogStartOSD.Checked then LogAction('Not supported: START_OSD',True);

        MDAPI_OSD_DRAWBLOCK:
            if LogOSDDrawBlock.Checked then LogAction('Not supported: OSD_DRAWBLOCK',True);

        MDAPI_OSD_SETFONT:
            if LogOSDSetFont.Checked then LogAction('Not supported: OSD_SETFONT',True);

        MDAPI_OSD_TEXT:
            if LogOSDText.Checked then LogAction('Not supported: OSD_TEXT',True);

        MDAPI_SEND_OSD_KEY:
            if LogSendOSDKey.Checked then LogAction('Not supported: OSD_KEY',True);

        MDAPI_STOP_OSD:
            if LogStopOSD.Checked then LogAction('Not supported: STOP_OSD',True);

        PROGAPI_GET_SIGNAL_LEVEL:
            begin
                if LogPROGAPIGetSignalLevel.Checked then LogAction('PROGAPI_GET_SIGNAL_LEVEL');
                DVBGetTunerInfo(@DVBTunInfo);
                ProgAPISignalState:=PProgAPISignalState(msg.LParam);
                ProgAPISignalState^.dwLevel:=DVBTunInfo.Level;
                ProgAPISignalState^.dwQuality:=DVBTunInfo.Quality;
                msg.Result:=1;
                exit;
            end;

        PROGAPI_SET_CHANNEL:
            begin
                ProgAPIGetChannelName:=PProgAPIGetChannelName(msg.LParam);
                n:=ProgAPIGetChannelName.dwSID;
                if LogPROGAPISetChannelName.Checked then LogAction('PROGAPI_SET_CHANNEL: SID='+inttohex(n,4)+' TID='+inttohex(ProgAPIGetChannelName.dwTID,4));

                if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
                else i:=GET_TRYFROMCACHE;
                DVBParams(@TPInfo, GET_ALLINFO or i);
                for i:=0 to TPInfo.wChCount-1 do if n=TPInfo.Channel[i].wSID then begin
                    get_name(@TPInfo.Channel[i].Ch_Name, @ProgAPIGetChannelName^.ChannelName,32);
                    get_name(@TPInfo.Channel[i].Ch_Name, @ProgAPIGetChannelName^.ChannelCaption,128);
                    get_name(@TPInfo.Channel[i].Prov_Name, @ProgAPIGetChannelName^.ProviderName,32);
//                    DVBDSDisplay(@DSDisplay,DVB_DSDISP_GETSTATE);
//                    if (DSDisplay.State=DVB_DSDISP_PLAY) then
                        k:=SET_STARTCAPTURE;
//                    else k:=SET_STOPCAPTURE;
                    TPInfo.wChCurrent:=i;
                    DVBParams(@TPInfo, SET_COPYALLDATA or SET_TUNER or SET_PIDS or k);
                    msg.Result:=1;
                    exit;
                end;
            end;

        PROGAPI_GET_CHANNEL_NAME:
            begin
                ProgAPIGetChannelName:=PProgAPIGetChannelName(msg.LParam);
                n:=ProgAPIGetChannelName.dwSID;
                if LogPROGAPIGetChannelName.Checked then LogAction('PROGAPI_GET_CHANNEL_NAME: SID='+inttohex(n,4)+' TID='+inttohex(ProgAPIGetChannelName.dwTID,4));

                if ChannelInfoOnlyFromCache.Checked then i:=GET_ONLYFROMCACHE
                else i:=GET_TRYFROMCACHE;
                DVBParams(@TPInfo, GET_ALLINFO or i);
                for i:=0 to TPInfo.wChCount-1 do if n=TPInfo.Channel[i].wSID then begin
                    get_name(@TPInfo.Channel[i].Ch_Name, @ProgAPIGetChannelName^.ChannelName,32);
                    get_name(@TPInfo.Channel[i].Ch_Name, @ProgAPIGetChannelName^.ChannelCaption,128);
                    get_name(@TPInfo.Channel[i].Prov_Name, @ProgAPIGetChannelName^.ProviderName,32);
                    msg.Result:=1;
                    exit;
                end;
            end;

        MDAPI_DVB_COMMAND:
            begin
                if PDVB_COMMAND(msg.LParam)^.Cmd_laenge<>7 then begin
                    ShowMessage('DVB_COMMAND: Unknow command length='+inttostr(PDVB_COMMAND(msg.LParam)^.Cmd_laenge));
                    msg.Result:=0;
                end else begin
                    for i:=4 to 13 do SAA_COMMAND[i]:=PDVB_COMMAND(msg.LParam)^.Cmd_Buffer[i];
                    DVBSendSAACmd(@SAA_COMMAND[0],7);
                    msg.Result:=1;
                end;
                exit;
            end;

    else LogAction('Unknown API command: '+IntToHex(msg.WParam,8),True);
    end;

    msg.Result:=0;
end;

procedure CMD_Filter(Param: DWORD; COMMAND: Cardinal; hFilter: DWORD);stdcall;
var
    i:integer;
    s:string;
begin
    if MDPluginWorkspace.LogDVBCoreCOMMAND.Checked then LogAction('DVBCore COMMAND : '+inttoHex(COMMAND,8));
    case COMMAND of

        FT_CMD_CHANNEL_WILLBE_CHANGED,
        FT_CMD_TRANSPONDER_WILLBE_CHANGED,
//        FT_CMD_TRANSPONDER_CHANGED,
//        FT_CMD_DISEQC_WILLBE_CHANGED,
//        FT_CMD_DISEQC_CHANGED,
//        FT_CMD_USER:

        FT_CMD_CLOSE,
        FT_CMD_DISABLE_FILTERS:
            MDPluginWorkspace.CloseFilters;

        FT_CMD_ENABLE_FILTERS,
        FT_CMD_CHANNEL_CHANGED:
           begin
            MDPluginWorkspace.ChangeChannel;
           end;

        FT_CMD_NEW_CK:
            with MDPluginWorkspace do begin
                if LogDVBCommand.Checked then begin
                    s:='';
                    for i:=0 to 13 do s:=s+' '+IntToHex(PByteArray(Param)[i],2);
                    LogAction('DVB_COMMAND: '+s);
                end;
            end;
    end;
end;

procedure TMDPluginWorkspace.FormCreate(Sender: TObject);
//var
//    i: integer;

    procedure ProcessMenu(menu:TMenuItem);
    var i:integer;
        s:string;
    begin
      for i:=0 to Menu.Count-1 do
            if Menu.Items[i].Count>0 then ProcessMenu(Menu.Items[i])
            else begin
              if Menu.Items[i].Checked then s:='1'
              else s:='0';
              s:=ini.ReadString('MD API Wrapper', Menu.Items[i].Name, s);
              if s='1' then Menu.Items[i].Checked:=True
              else Menu.Items[i].Checked:=False;
            end;
      for i:=1 to 5 do
        if ((FindComponent('P'+IntToStr(i)) as TMenuItem).Checked=True) then begin
          NEW_BUFFER_SIZE:=184*i+1;
          break;
        end;
    end;

begin
//    for i:=0 to MaxFilterCount do Filters[i].Active:=False;
//    for i:=0 to MaxPluginCount do MDPlugins[i].DLLID:=0;

    InitializeCriticalSection(CAT_CS);
    CATEvent:=CreateEvent(nil, true, false, nil);
    ini:=TIniFile.Create(extractfilepath(application.ExeName)+'Settings.ini');
    ProcessMenu(OptionsMenu);
    InitMDPlugins;
    Top:=0;
    Left:=0;
end;

function DVBPLGInit(CardType: DWORD; AppWindow: HWND; ExtraInfo: Pointer=nil): BOOL;stdcall;
begin
    if not inited then DVBAddFilter(HOOK_COMMANDS,@CMD_Filter);
    inited:=true;
    result:=true;
    SetWindowLong(MDPluginWorkSpace.Handle, GWL_HWNDPARENT, AppWindow);
end;

function  DVBPLGCall(Code: DWORD =0; ExtraInfo: Pointer = nil): BOOL; stdcall;
begin
    MDPluginWorkspace.Show;
    result:=true;
end;

function DVBPLGClose(Reason: DWORD =0): BOOL;stdcall;

    procedure ProcessMenu(menu:TMenuItem);
    var
        i:integer;
    begin
        for i:=0 to Menu.Count-1 do
            if Menu.Items[i].Count>0 then ProcessMenu(Menu.Items[i])
                else if Menu.Items[i].Checked then MDPluginWorkspace.ini.WriteString('MD API Wrapper', Menu.Items[i].Name, '1')
                        else MDPluginWorkspace.ini.WriteString('MD API Wrapper', Menu.Items[i].Name, '0')
    end;
var
    i: integer;
begin
    MDPluginWorkspace.Close;
    ProcessMenu(MDPluginWorkspace.OptionsMenu);
    Initialized:=False;
    MDPluginWorkspace.CloseFilters;
    for i:=0 to PluginCount-1 do
        if (MDPlugins[i].DLLID<>0) then begin
            try
               if Assigned(MDPlugins[i].ExitMDPlugin) then
                   MDPlugins[i].ExitMDPlugin(HInstance,MDPluginWorkspace.Handle,true);
            except
            end;
        end;
//    MDPluginWorkspace.Destroy;
    result:=true;
end;

procedure TMDPluginWorkspace.FormDestroy(Sender: TObject);

var
    i: integer;
begin
    ini.Free;
    for i:=0 to PluginCount-1 do
        if (MDPlugins[i].DLLID<>0) then
               FreeLibrary(MDPlugins[i].DLLID);
    CloseHandle(CATEvent);
    DeleteCriticalSection(CAT_CS);
end;

function DVBPLGGetInfo(PLG_Info: pPLG_INFO): BOOL; stdcall;
begin
    PLG_Info.hdrVer:=100;
    PLG_Info.dwShortCut:=TextToShortCut('Ctrl+M');
    PLG_Info.dwTTL:= 0; //PLG_TTL_FOREVER;
    PLG_Info.Name:='MD API Wrapper';
    result:=true;
end;

procedure TMDPluginWorkspace.PluginLogKeyPress(Sender: TObject;
  var Key: Char);
begin
    if Key=#27 then MDPluginWorkspace.Hide;
end;

procedure TMDPluginWorkspace.ViewAllFilters(Sender: TObject);
var
    i,k:integer;
begin
    k:=0;
    for i:=0 to MaxFilterCount do if Filters[i].Active then begin
        LogAction('----------------------------');
        LogAction('Filter'+inttostr(i+1)+' Name      : '+Filters[i].Name);
        LogAction('Filter'+inttostr(i+1)+' PID       : '+inttoHex(Filters[i].PID,8));
        LogAction('Filter'+inttostr(i+1)+' Handle    : '+inttoHex(Filters[i].Handle,8));
        LogAction('Filter'+inttostr(i+1)+' Proc      : '+inttoHex(Cardinal(@Filters[i].Proc),8));
        LogAction('Filter'+inttostr(i+1)+' DLL_ID    : '+inttoHex(Filters[i].DLL_ID,8));
        LogAction('Filter'+inttostr(i+1)+' Filter_ID : '+inttoHex(Filters[i].Filter_ID,8));
        LogAction('Filter'+inttostr(i+1)+' BufLen    : '+inttoHex(Filters[i].BufLen,8));
        LogAction('Filter'+inttostr(i+1)+' Buffer    : '+inttoHex(Cardinal(Filters[i].Buffer),8));
        LogAction('----------------------------');
        k:=1;
    end;
    if (k<>1) then LogAction('Filters is not found');
end;

procedure TMDPluginWorkspace.ViewAllPlugins(Sender: TObject);
var
    i,k:integer;
begin
    k:=0;
    for i:=0 to MaxPluginCount do if (MDPlugins[i].DLLID<>0) then begin
        LogAction('-------------------------------------------');
        LogAction('Plugin'+inttostr(i+1)+' Name                     : '+MDPlugins[i].Name);
        LogAction('Plugin'+inttostr(i+1)+' HotKey                   : '+MDPlugins[i].HotKey);
        LogAction('Plugin'+inttostr(i+1)+' DLLID                    : '+inttoHex(MDPlugins[i].DLLID,8));
        LogAction('Plugin'+inttostr(i+1)+' PluginMenu               : '+inttoHex(MDPlugins[i].PluginMenu,8));
        LogAction('Plugin'+inttostr(i+1)+' InitPlugin               : '+inttoHex(Cardinal(@MDPlugins[i].InitPlugin),8));
        LogAction('Plugin'+inttostr(i+1)+' GetPluginName            : '+inttoHex(Cardinal(@MDPlugins[i].GetPluginName),8));
        LogAction('Plugin'+inttostr(i+1)+' ProcessPluginMenuCommand : '+inttoHex(Cardinal(@MDPlugins[i].ProcessPluginMenuCommand),8));
        LogAction('Plugin'+inttostr(i+1)+' ProcessChannelChange     : '+inttoHex(Cardinal(@MDPlugins[i].ProcessChannelChange),8));
        LogAction('Plugin'+inttostr(i+1)+' ExitMDPlugin             : '+inttoHex(Cardinal(@MDPlugins[i].ExitMDPlugin),8));
        LogAction('Plugin'+inttostr(i+1)+' ProcessHotKey            : '+inttoHex(Cardinal(@MDPlugins[i].ProcessHotKey),8));
        LogAction('Plugin'+inttostr(i+1)+' ProcessFilterClose       : '+inttoHex(Cardinal(@MDPlugins[i].ProcessFilterClose),8));
        LogAction('Plugin'+inttostr(i+1)+' ProcessRecPlay           : '+inttoHex(Cardinal(@MDPlugins[i].ProcessRecPlay),8));
        LogAction('Plugin'+inttostr(i+1)+' ProcessOsdKey            : '+inttoHex(Cardinal(@MDPlugins[i].ProcessOsdKey),8));
        LogAction('-------------------------------------------');
        k:=1;
    end;
    if (k<>1) then LogAction('Plugins is not found');
end;

procedure TMDPluginWorkspace.BufferSizeChange(Sender: TObject);
begin
    NEW_BUFFER_SIZE:=(184*(Sender as TMenuItem).Tag)+1;
end;

end.
